/*
 * SonarQube
 * Copyright (C) 2009-2022 SonarSource SA
 * mailto:info AT sonarsource DOT com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
package org.sonar.ce.task.projectanalysis.pushevent;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import java.util.LinkedList;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import org.jetbrains.annotations.NotNull;
import org.sonar.api.ce.ComputeEngineSide;
import org.sonar.ce.task.projectanalysis.analysis.AnalysisMetadataHolder;
import org.sonar.ce.task.projectanalysis.component.Component;
import org.sonar.ce.task.projectanalysis.component.TreeRootHolder;
import org.sonar.core.issue.DefaultIssue;
import org.sonar.db.protobuf.DbCommons;
import org.sonar.db.protobuf.DbIssues;
import org.sonar.db.pushevent.PushEventDto;
import org.sonar.server.issue.TaintChecker;

import static java.nio.charset.StandardCharsets.UTF_8;
import static java.util.Objects.requireNonNull;
import static java.util.Objects.requireNonNullElse;

@ComputeEngineSide
public class PushEventFactory {
  private static final Gson GSON = new GsonBuilder().create();

  private final TreeRootHolder treeRootHolder;
  private final AnalysisMetadataHolder analysisMetadataHolder;
  private final TaintChecker taintChecker;

  public PushEventFactory(TreeRootHolder treeRootHolder,
    AnalysisMetadataHolder analysisMetadataHolder, TaintChecker taintChecker) {
    this.treeRootHolder = treeRootHolder;
    this.analysisMetadataHolder = analysisMetadataHolder;
    this.taintChecker = taintChecker;
  }

  public Optional<PushEventDto> raiseEventOnIssue(DefaultIssue currentIssue) {
    var currentIssueComponentUuid = currentIssue.componentUuid();
    if (!taintChecker.isTaintVulnerability(currentIssue) || currentIssueComponentUuid == null) {
      return Optional.empty();
    }

    var component = treeRootHolder.getComponentByUuid(Objects.requireNonNull(currentIssue.componentUuid()));
    if (currentIssue.isNew() || currentIssue.isCopied() || isReopened(currentIssue)) {
      return Optional.of(raiseTaintVulnerabilityRaisedEvent(component, currentIssue));
    }
    if (currentIssue.isBeingClosed()) {
      return Optional.of(raiseTaintVulnerabilityClosedEvent(currentIssue));
    }
    return Optional.empty();
  }

  private static boolean isReopened(DefaultIssue currentIssue) {
    var currentChange = currentIssue.currentChange();
    if (currentChange == null) {
      return false;
    }
    var status = currentChange.get("status");
    return status != null && status.toString().equals("CLOSED|OPEN");
  }

  private PushEventDto raiseTaintVulnerabilityRaisedEvent(Component component, DefaultIssue issue) {
    TaintVulnerabilityRaised event = prepareEvent(component, issue);
    return new PushEventDto()
      .setName("TaintVulnerabilityRaised")
      .setProjectUuid(treeRootHolder.getRoot().getUuid())
      .setPayload(serializeEvent(event));
  }

  private TaintVulnerabilityRaised prepareEvent(Component component, DefaultIssue issue) {
    TaintVulnerabilityRaised event = new TaintVulnerabilityRaised();
    event.setProjectKey(issue.projectKey());
    event.setCreationDate(issue.creationDate().getTime());
    event.setKey(issue.key());
    event.setSeverity(issue.severity());
    event.setRuleKey(issue.getRuleKey().toString());
    event.setType(issue.type().name());

    event.setBranch(analysisMetadataHolder.getBranch().getName());

    DbIssues.Locations issueLocations = requireNonNull(issue.getLocations());

    Location mainLocation = new Location();
    mainLocation.setMessage(issue.getMessage());

    mainLocation.setFilePath(component.getName());

    TextRange mainLocationTextRange = getTextRange(issueLocations.getTextRange(), issueLocations.getChecksum());
    mainLocation.setTextRange(mainLocationTextRange);
    event.setMainLocation(mainLocation);

    List<Flow> flows = new LinkedList<>();
    for (DbIssues.Flow sourceFlow : issueLocations.getFlowList()) {
      Flow flow = new Flow();
      List<Location> locations = new LinkedList<>();
      for (DbIssues.Location sourceLocation : sourceFlow.getLocationList()) {
        Location location = new Location();
        var locationComponent = treeRootHolder.getComponentByUuid(sourceLocation.getComponentId());
        location.setFilePath(requireNonNullElse(locationComponent, component).getName());
        location.setMessage(sourceLocation.getMsg());

        TextRange textRange = getTextRange(sourceLocation.getTextRange(), sourceLocation.getChecksum());
        location.setTextRange(textRange);

        locations.add(location);
      }
      flow.setLocations(locations);
      flows.add(flow);
    }
    event.setFlows(flows);
    return event;
  }

  private PushEventDto raiseTaintVulnerabilityClosedEvent(DefaultIssue issue) {
    TaintVulnerabilityClosed event = new TaintVulnerabilityClosed(issue.key(), issue.projectKey());
    return new PushEventDto()
      .setName("TaintVulnerabilityClosed")
      .setProjectUuid(treeRootHolder.getRoot().getUuid())
      .setPayload(serializeEvent(event));
  }

  private static byte[] serializeEvent(Object event) {
    return GSON.toJson(event).getBytes(UTF_8);
  }

  @NotNull
  private static TextRange getTextRange(DbCommons.TextRange source, String checksum) {
    TextRange textRange = new TextRange();
    textRange.setStartLine(source.getStartLine());
    textRange.setStartLineOffset(source.getStartOffset());
    textRange.setEndLine(source.getEndLine());
    textRange.setEndLineOffset(source.getEndOffset());
    textRange.setHash(checksum);
    return textRange;
  }

}
